"""
Sample code for Huiwen Jia and Siqian Shen, "Benders Cut Classification via Support Vector Machines for Solving Two-stage Stochastic Programs", INFORMS Jounal on Optimization, 2020.

@author: hwjia@umich.edu
"""
from gurobipy import *
import random
import xlwt
import numpy as np
import time
import math
import pandas as pd
from sklearn import svm
import xlrd
    
def learn_bd(Warehouse, setup, capacity, Factory, lostsalecost, Pair, cost , Scenario, probability, Demand, gapv, K, N, training_fileposition, fileposition, key, key_low, key_change_time, itelimit, pretrain,timelimit):
    
    '''
    Stage 1
    '''
    print("##########################################")
    print("Learn-Stage1")
    print("##########################################")
    nsce=len(probability)
        
    if pretrain == 0:        
        workbookdata = xlwt.Workbook() 
        sheetdata = workbookdata.add_sheet('TrainingData',cell_overwrite_ok=True)
        sheetdata.write(0,0,'Index')
        sheetdata.write(0,1,'Current Violation')
        sheetdata.write(0,2,'# of previous cuts')
        sheetdata.write(0,3,'obj change')
        
        I=0
        dataindex=1
        mt_v=[]
        st_v=[]
        while I<K:      
            I += 1 # 
            np.random.seed(1234+I)
            m1 = Model("MP") 
            m1.setParam( 'OutputFlag', False )
            x = m1.addVars(Warehouse, vtype=GRB.BINARY, name="x")
            z = m1.addVars(Scenario,vtype=GRB.CONTINUOUS, name="z")
            obj = LinExpr();
            obj += x.prod(setup);
            obj += z.prod(probability);
            m1.setObjective(obj, GRB.MINIMIZE);
            
       #     print("\n Solution of initial Master Problem \n \n")
            
            mastertime=0
            start = time.clock()
            m1.optimize()
            end = time.clock()
            mastertime=mastertime + (end - start)
            zvalue = m1.getAttr('x', z)
            xvalue = m1.getAttr('x', x)
                
            # Begin iteration
            cutssce = {}  # record number of cuts generated by each scenario
            zup={}
            for s in Scenario:
                cutssce[s] = 0
                zup[s] = math.inf
         #  print("\n Begin iteration")
            t=0 # index of iteration
            sheet1row=2
            sheet1column=3
            lowerbound = m1.objVal
            upperbound = (xvalue.prod(setup) + quicksum(zup[s]*probability[s] for s in Scenario)).getValue()
            dif = upperbound - lowerbound
            subtime=0
            while (t<N) :
                t=t+1 
                newcut = 0
                Scenario = tuplelist(Scenario)
                SamSeq = random.sample(range(nsce),nsce)
                ns=0
                while (ns < nsce and newcut < 1):  
                    s = Scenario[SamSeq[ns]]
                    ns += 1   
                    ms = Model("SP")
                    ms.setParam( 'OutputFlag', False )
                    beta = ms.addVars(Factory, vtype=GRB.CONTINUOUS, name="beta") #lostsale
                    y = ms.addVars(Pair, vtype=GRB.CONTINUOUS, name="y")
                    obj = LinExpr();
                    obj += beta.prod(lostsalecost);
                    obj += y.prod(cost);
                    ms.setObjective(obj, GRB.MINIMIZE);
                    ms.addConstrs( (Demand[s,j]- y.sum('*',j)  <= beta[j] for j in Factory), "lostsale")
                    ms.addConstrs( ( y.sum(i,'*') <= capacity[i]*xvalue[i] for i in Warehouse ), "Cap_Cons")
                    start = time.clock()
                    ms.optimize()
                    end = time.clock()
                    subtime=subtime + (end - start)
                    sobj = ms.objVal
                    if sobj > zvalue[s] :
                        newcut = 1
                        sheetdata.write(dataindex,0,t)
                        sheetdata.write(dataindex,1,sobj - zvalue[s])
                        sheetdata.write(dataindex,2,cutssce[s])
                        sheetdata.write(dataindex,3,0)
                        cutssce[s] = cutssce[s] + 1
                        zup[s] = sobj
                        sdual = ms.getAttr('Pi') 
                        l=0
                        pii = {}
                        for j in Factory:
                            pii[j] = sdual[l]
                            l=l+1 
                        hi = {}
                        for i in Warehouse:
                            hi[i] = sdual[l]
                            l=l+1  
                        currentS = [{}]
                        currentS = [s]
                        name = '_'.join(["Cut generated in Iteration",str(t),"Scenario",s]) 
                        #print("\n  \n !!! New cut -------- %s",  name) 
                        m1.addConstr( 
                        (z[s] >= quicksum( capacity[i]*x[i]*hi[i] for i in Warehouse)  + quicksum( Demand[s,j] * ( -  pii[j])  for j in Factory) ) , name) 
                       #print("Immediately add this cut into Master Problem")  
                       
                if newcut==1:
                    objbefore = m1.objVal
                    start = time.clock()
                    m1.optimize()
                    end = time.clock()
                    mastertime=mastertime + (end - start)
                    
                    lowerbound = m1.objVal
                    if lowerbound==0:
                        lowerbound=0.1
                    upperbound = min((xvalue.prod(setup) + quicksum(zup[s]*probability[s] for s in Scenario)).getValue(),upperbound)
                    dif = (upperbound - lowerbound)/lowerbound
                    objchange = lowerbound - objbefore
                    sheetdata.write(dataindex,3,objchange)
                    dataindex += 1
                    xvalue = m1.getAttr('x', x)
                    zvalue = m1.getAttr('x', z)
                else:
                    break
        

            mt_v.append(mastertime)
            st_v.append(subtime)
            print('K=',I)
            print('mastertime-',mastertime)
            print('subtime-',subtime)
        
        filename = ''.join([fileposition,"train_data.xls"]) 
        workbookdata.save(filename)         
        
        t_D = {'master':mt_v, 'sub':st_v}
        t_D = pd.DataFrame(t_D)  
        t_D.to_excel("".join([fileposition,'phase1_time.xlsx'])) 
        

    '''
    Stage 2
    '''   
    filename = ''.join([training_fileposition,"train_data.xls"])
    print("##########################################")
    print("Learn-Stage2")
    print("##########################################")     

    
    violate=[]
    cutnumber=[]
    label=[]
     
    data = xlrd.open_workbook(filename)
    table = data.sheets()[0] 
    violate = violate + table.col_values(1)[1:]
    cutnumber =cutnumber +  table.col_values(2)[1:]
    newlabel = table.col_values(3)[1:]
    for npath in range(0, K):
        startpos = npath*N
        for n in range(startpos+0, startpos+N-1) :
            if newlabel[n]<=(key*newlabel[n+1]) :
                newlabel[n] = -1
            else:
                newlabel[n] = 1
        newlabel[startpos+N-1] = 1
    label = label + newlabel
    
    
    training_D = {'vio':violate, 'ncut':cutnumber, 'label':label}
    training_D = pd.DataFrame(training_D)
    training_D.drop_duplicates(['vio','ncut'], inplace=True) 
    
    violate=list(training_D['vio'])
    cutnumber=list(training_D['ncut'])
    label=list(training_D['label'])
    
    trainx = list(zip(violate, cutnumber))
    
    clf=svm.SVC()  
    clf.fit(trainx,label)


    '''
    Stage 3
    ''' 
    print("##########################################")
    print("Learn-Stage3")
    print("##########################################")
    totaltime_start=time.clock()
    m1 = Model("MP")   
    m1.setParam( 'OutputFlag', False )
    x = m1.addVars(Warehouse, vtype=GRB.BINARY, name="x")
    z = m1.addVars(Scenario,vtype=GRB.CONTINUOUS, name="z")
    obj = LinExpr();
    obj += x.prod(setup);
    obj += z.prod(probability);
    m1.setObjective(obj, GRB.MINIMIZE);
    
    retrain_iteration=[]
    trainingtime_v=[]
    mastertime=0
    start = time.clock()
    m1.optimize()
    end = time.clock()
    mastertime=mastertime + (end - start)
    zvalue = m1.getAttr('x', z)
    xvalue = m1.getAttr('x', x)
    
    cutssce = {}  # used to record number of cuts generated by each scenario
    zup={}
    for s in Scenario:
        cutssce[s] = 0
        zup[s] = math.inf


    lowerbound = m1.objVal
    upperbound = (xvalue.prod(setup) + quicksum(zup[s]*probability[s] for s in Scenario)).getValue()
    dif = upperbound - lowerbound
    subtime=0
    t=1 # index of iteration
    mastertimev=[]
    subtimev=[]
    ratiobindv=[]
    iterationv=[]
    realgapv=[]
    cutnv=[]
    lowerboundv=[]
    bd_flag=0 # using learnBD if ==1, switch to traditional BD
    pos_bd_flag = - math.inf
    gaplimit = gapv[len(gapv)-1]
    key_range = key - key_low
    
    pos_bd_flag = []

    while ( dif > gaplimit) :

        print("Iteration %d" %t)

        ncut=0
        for s in Scenario:  # for each subproblem

            ms = Model("SP")
            ms.setParam( 'OutputFlag', False )
            beta = ms.addVars(Factory, vtype=GRB.CONTINUOUS, name="beta") #lostsale
            y = ms.addVars(Pair, vtype=GRB.CONTINUOUS, name="y")
            obj = LinExpr();
            obj += beta.prod(lostsalecost);
            obj += y.prod(cost);
            ms.setObjective(obj, GRB.MINIMIZE);
            ms.addConstrs( (Demand[s,j]- y.sum('*',j)  <= beta[j] for j in Factory), "lostsale")
            ms.addConstrs( ( y.sum(i,'*') <= capacity[i]*xvalue[i] for i in Warehouse ), "Cap_Cons")
            start = time.clock()
            ms.optimize()
            end = time.clock()
            subtime=subtime + (end - start)
            sobj = ms.objVal
            testx = [(sobj - zvalue[s],cutssce[s])]
            if sobj > zvalue[s] :
                zup[s] = sobj
                if bd_flag == 0:
                    score = int(clf.predict(np.array(testx).reshape(1, -1)))
                else:
                    score = 1
                    
                if score == 1 : 
                   cutssce[s] = cutssce[s] + 1
                   sdual = ms.getAttr('Pi') 
                   l=0
                   pii = {}
                   for j in Factory:
                       pii[j] = sdual[l]
                       l=l+1 
                   hi = {}
                   for i in Warehouse:
                       hi[i] = sdual[l]
                       l=l+1 
                   name = '_'.join(["Cut generated in Iteration",str(t),"Scenario",s])   
                   ncut += 1
                   m1.addConstr( 
                   (z[s] >=   quicksum( capacity[i]*x[i]*hi[i] for i in Warehouse)  + quicksum( Demand[s,j] * ( -  pii[j])  for j in Factory) ) , name) 
        pos_bd_flag.append(key)
        if ncut==0 and bd_flag==0:
            if key>key_low:
                key=key - (key_range)/key_change_time
                clf, trainingtime =train_svm(training_fileposition,fileposition, key, t, K, N)
                retrain_iteration.append(t)
                trainingtime_v.append(trainingtime)
            else:
                bd_flag = 1    
                key='tradi'
                trainingtime_v.append(0)
                print("##########################################")
                print("TraBD----Iteration------", t)
                print("##########################################") 
        else:
            trainingtime_v.append(0)
            start = time.clock()
            m1.optimize()
            end = time.clock()
            mastertime=mastertime + (end - start)           
            lowerbound = m1.objVal
            if lowerbound==0:
                lowerbound=0.1
            upperbound = min((xvalue.prod(setup) + quicksum(zup[s]*probability[s] for s in Scenario)).getValue(),upperbound)
            dif = (upperbound - lowerbound)/lowerbound
            print("UB",upperbound,"     LB",lowerbound,"    dif",dif, '   mastertime', mastertime)
            xvalue = m1.getAttr('x', x)
            zvalue = m1.getAttr('x', z)
        t=t+1
        sla=m1.getAttr("Slack") 
        l=0
        nbind=0
        while l < len(sla) :
            if abs(sla[l])<=0.00000001:
                nbind += 1
            l += 1
        if len(sla)== 0 :
            ratiobind = math.inf
        else:
            ratiobind = (nbind/len(sla))
        cutnv.append(len(sla))
        mastertimev.append(mastertime)
        subtimev.append(subtime)
        iterationv.append(t-1)
        ratiobindv.append(ratiobind)
        realgapv.append( dif )
        lowerboundv.append(lowerbound)
       # if t>itelimit:
        #    break
        if mastertime>timelimit:
            break
    print("##########################################")    
    print('retrain---iterations----', retrain_iteration)
    print("##########################################")
    return (lowerboundv, 
            realgapv,
            iterationv,
            cutnv,
            ratiobindv,
            mastertimev,
            subtimev,
            pos_bd_flag,
            trainingtime_v
            )


def train_svm(train_fileposition,fileposition, key, t, K, N):
    filename = ''.join([train_fileposition,"train_data.xls"])
    print("##########################################")
    print("Training----Iteration------", t)
    print('key-update', key)
    print("##########################################")     

    
    violate=[]
    cutnumber=[]
    label=[]
     
    data = xlrd.open_workbook(filename)
    table = data.sheets()[0] 
    violate = violate + table.col_values(1)[1:]
    cutnumber =cutnumber +  table.col_values(2)[1:]
    newlabel = table.col_values(3)[1:]
    for npath in range(0, K):
        startpos = npath*N
        for n in range(startpos+0, startpos+N-1) :
            if newlabel[n]<=(key*newlabel[n+1]) :
                newlabel[n] = -1
            else:
                newlabel[n] = 1
        newlabel[startpos+N-1] = 1
    label = label + newlabel
    
    
    training_D = {'vio':violate, 'ncut':cutnumber, 'label':label}
    training_D = pd.DataFrame(training_D)
    training_D.drop_duplicates(['vio','ncut'], inplace=True) 
    
    violate=list(training_D['vio'])
    cutnumber=list(training_D['ncut'])
    label=list(training_D['label'])
    
    trainx = list(zip(violate, cutnumber))
    
    
    clf=svm.SVC()  
    start = time.clock()
    clf.fit(trainx,label)
    end = time.clock()
    trainingtime=(end - start) 
    print("!! -- Phase 2 training time", trainingtime)
    
    return(clf,trainingtime)